//Chris Carlson.  Final Project 220 b.
//Multiband feedback delay spatializer

//signal path:

//input signal (connected externally) =>
//dry and wet gains.
//wet gain =>  
//Bandpass Filterbank[3] => 
//N*F delays (N = number of channels, F = num bandpass filters) =>
//N*F feedback gain objects => 
//N*F lowpass filters =>
//previous delay inputs (feedback loop)
//each delay out to an envelope to a single audio channel

//In the current implementation each output channel contains all three input filters
//delayed by unique amounts (compared to other channels).  So, for 8 channels and 3 input filters, we have 24 total
//delay lines/envelopes/etc. 

//Other nice features:  
//filter stability monitoring shred to make sure filters that are about to blow up get silenced  
//filter cutoff slewing functions (sinusoidally vary bandpass cutoffs and feedback delay filter cutoffs)
//spatialization method allows for circular motion or linear jumps between speakers.  


public class FilterSpat{
	
	//preloaded sound playback optional
	//optional sound to play through 
	//SndBuf buf;
	
	//"/Users/carlson/Desktop/Sketchbook/" => string soundDir;
	
	//soundDir + "RE_201.wav" => buf.read;
	
	
	//set number of output channels/speakers
	dac.channels() => static int numChans;
	//8 => static int numChans;
	
	//connect dry output to all dac channels
	Gain dry => dac;
	//optional preloaded sound - dry output
	//buf => dry
	0.5 => dry.gain;

	//wet input gain  (pre filter, envelopes)
	Gain wet;
	
	//number of filters to use
	3 => int numFilters; //6 for 4 chan
	
	//filter objects
	BPF filters[numFilters];
	
	//total number of outputs
	numFilters*numChans => int totalOuts;
	
	//print total number of outputs
	<<<totalOuts + "total">>>;
	
	
	//parameters for automated frequency assignment (replaced
	//with hard coded values (tuned in the CCRMA stage)
	//60 => float bottomFreq;
	//1000  => float freqMult;  //900 for 4 chan
	
	
	//tailor filters for guitar
	100 => filters[0].freq;
	0.35 => filters[0].Q;
	1500 => filters[1].freq;
	0.4 => filters[1].Q;
	9000 => filters[2].freq;
	0.6 => filters[2].Q;
	
	//connect wet input to the bandpass filters and set gains
	for (int i; i < numFilters; i++){
		//automated filter freq + Q assignment. (replaced by hard coded vals)
		//freqMult*(i*i) + bottomFreq => filters[i].freq;
		//0.9 => filters[i].Q;
		
		wet => filters[i];
		
		//optional assignment of prerecorded sound to filters
		//if (i > 2 && i < numFilters -2){
		//	buf => filters[i];
		//}
	
		0.95 => filters[i].gain;
	}
	
	
	//assign initial filter channels (corresponding to envelope output)
	//the nth value of this array tells the current active (on) envelope number
	int filtChans[numFilters];
	for (int i; i < numFilters; i++){
		Std.rand2(0,numChans-1) + i*numChans => filtChans[i];
		//<<<"filtChans" + filtChans[i]>>>;
	}
	
	
	//set up channel envelopes
	Envelope envs[totalOuts];
	
	
	//initial envelope duration (for all envelopes)
	1::second => dur initDur;
	
	
	//Filtered feedback delay loop objects
	DelayA d[totalOuts];
	Gain fbks[totalOuts];
	LPF lfs[totalOuts];
	
	//set envelope durations, connect to dac, connect filters to appropriate envelopes
	//note that we have one envelope per channel, per filter. 
	BPF connectFilter;
	//counter for filter assignment changes
	0 => int cvar;
	for (int i; i < totalOuts; i++){
		//duration set
		envs[i].duration(initDur);
		envs[i] => dac.chan(i % numChans);
		//grab appropriate filter reference if last filter has been assigned to envelopes connecting to all output channels
		if ((i % numChans) == 0){
			filters[cvar] @=> connectFilter;
			cvar++;
		}
		
		//connect new filter to envelope
		connectFilter => d[i] =>fbks[i] => lfs[i] => d[i] => envs[i];
		
		//set delay and feedback filter parameters
		1.5::second => d[i].max;
		Std.rand2f(0.020,0.35)::second => d[i].delay;
		if (i < (.75*totalOuts)){
			75*(i+1) => lfs[i].freq;
			Std.rand2f(0.85,0.93) => fbks[i].gain;
			
		}else{
			900 + 35*(i+1) => lfs[i].freq;
			Std.rand2f(0.82,0.92) => fbks[i].gain;
			
		}
	}
	
	
	//optionally set soundbuf params
	//0 => buf.pos;
	//0.7 => buf.gain;
	//1 => buf.rate;
	//buf.loop(1);
	
	//constructor...may not be needed, but will keep here anyway
	fun void FilterSpat(){};
	
	//turn on the spatialization and slewing functions, etc (call externally)
	fun void run(){
		//turn all envelopes on initially
		for (int i; i < filtChans.size(); i++){
			envs[filtChans[i]].keyOn();
			//<<<filtChans[i]>>>;
		} 
		
		
		//counter variable for loop
		int counter[numFilters];
		1 => float maxDur;
		
		//spork slewing function for filter cutoffs
		spork~ slideCutoffs();
		
		//spork filter stability monitoring shred (prevent blowups by turning off envelopes before they occur)
		spork~ monitorStability();
		
		//loop/spatialize by turning on/of envelopes
		while (true){
			
			//move next round of filters through sound field using one of the following methods
			for (int i; i < numFilters;i++){
				if (maybe){
					if (maybe){
						counter[i]++;
						//pan circular
						move(i,(counter[i] % numChans) + i*numChans,Std.rand2f(0.1,maxDur)::second);
					}else{
						if (maybe){
							//random location switch
							move(i,Std.rand2(0,numChans-1) + i*numChans,Std.rand2f(0.1,maxDur)::second);
						}
					}
				}
				
			}
			
			//advance time stochastically
			Std.rand2f(0.01,0.7)::second => now;
			
		}
	}
	
	
	
	//spatialization method
	fun void move(int filtIdx,int newChan,dur fadeDur){
		//set fade time for old and new env
		envs[filtChans[filtIdx]].duration(fadeDur);
		envs[newChan].duration(fadeDur);
		
		//turn off old env (fade out)
		envs[filtChans[filtIdx]].keyOff();
		
		//turn on new env (fade in)
		envs[newChan].keyOn();
		
		//store new channel position to filter channel array (so we know which channel is active for a given filter)
		newChan => filtChans[filtIdx];
	}
	
	//move the cutoff frequencies of the delay line 
	//feedback filters and the channel filters
	fun void slideCutoffs(){
		0 => int r;
		300 => float width; //frequency shifting width for feedback delay filters
		100 => float smallWidth; // " for input bandpass filters
		while(true){
			for (int i; i < totalOuts; i++){
					Math.sin(r*pi/16)*width + lfs[i].freq() => float newFreq;
					if (newFreq < 0){
						10 => newFreq;
					}
					newFreq => lfs[i].freq;
			}
			for (int i; i < numFilters; i++){
				Math.sin(r*pi/16)*smallWidth + filters[i].freq() => float newFreq;
				if (newFreq < 0){
					0 => newFreq;
				}
				newFreq => filters[i].freq;
			}
			r++;
			Std.rand2f(0.3,0.7)::second => now;
		}
	}
	
	
	//turn all envelopes on or off (for big dynamic shifts)
	fun void allOnOff(int flag, dur duration){
		for (int i; i < envs.size(); i++){
			duration => envs[i].duration;
			if (flag == 1){
				envs[i].keyOn();
			}else{
				envs[i].keyOff();
			}
		}
		
		if (flag == 1){
			<<<"Filt ALL ON">>>;
		}else{
			<<<"Filt ALL OFF">>>;
		}
	}	
	
	
	//connect ugen to wet and dry inputs
	fun void connect(UGen g){
		g => dry;
		g => wet;
	}
	
	//provide number of output channels to querying object
	fun static int getNumChans(){
		return numChans;
	}
	
	
	
	//spork this to keep filters from blowing up.  will turn off envelope quickly if
	//envelope output exceeds threshold
 	fun void monitorStability(){
		while(true){
			for (int i; i < envs.size(); i++){
				if (envs[i].last() > 0.55){
					envs[i].duration(1::ms);
					envs[i].keyOff();
					<<<"STABILIZING: " + i>>>;
				}
			}
			1::ms=> now;
		}
	}	
}
