//Chris Carlson - 220b Final Project
//Live Looper using an array of LiSa objects
//Soft takeover implemented for gain values/controlled parameters (to prevent jumps in gain)
//this class also passes on control data to the filter spatializer effect

public class Looper{
	
	//declare signal path objects
	7 => int numLoops;
	LiSa loops[numLoops];
	Gain gains[numLoops];//individual loop gains
	Gain wetGains[numLoops]; //gains of loops sent to filter spatializer
	Envelope loopEnvs[numLoops];
	
	//keep track of solo state of each channel
	int solos[numLoops];
	//keep track of bidirectional looping mode of each channel
	int bis[numLoops];
	
	//dry looper signal to dac
	Gain dry => dac;
	//wet looper gain (wet = connected to filter spatializer)
	Gain wet;
	0.7 => wet.gain;
	0.7 => dry.gain;
	//direct output (input to dac)
	adc => Gain direct;
	0.8 => direct.gain;
	
	
	//min/max loop playback rates
	2 => float maxRate;
	-2 => float minRate;
	1 => float globalRate;
	
	//keep track of active voice numbers and loop durations
	int voices[numLoops];
	dur durs[numLoops];
	
	//initialize start time variable for loop
	time startTime;
	
	//soft takeover threshold (for midi control to avoid parameter jumps)
	0.2 => float sThresh; 
	0.2 => float rsThresh;
	
	//max duration of loop
	30::second => dur maxDur;
	
	//connect signal paths
	for(int i; i< numLoops; i++){
		//connect to loop inputs, envelopes, gains, dry output
		adc => loops[i] => loopEnvs[i] => gains[i] => dry;
		loopEnvs[i] => wetGains[i] => wet;
		loopEnvs[i].keyOn();
		loopEnvs[i].duration(20::ms);
		0 => wetGains[i].gain;//default = no signal sent to filter spat.
		0.7 => gains[i].gain;//default output of each loop
		//only one voice per lisa instance
		loops[i].maxVoices(1);
		//assign voice
		loops[i].getVoice() => voices[i];
		loops[i].duration(maxDur);
		//set fade recording fade in time
		loops[i].recRamp(50::ms);
		0::second => durs[i];
	}
	
	//instantiate filter spatializer object and connect sum of loop outputs (wet gain)
	//also connect direct guitar signal
	FilterSpat fs;
	fs.connect(wet);
	fs.connect(direct);
	
	//run filter spatializer
	spork ~fs.run();
	//<<<"init!">>>;
	
	//LOOPER METHODS
	
	//start recording new loop for voice number provided
	fun void recordNew(int vIdx){
		//<<<"recording to " + voices[vIdx]>>>;
		//fade out prevous loop and clear buffer
		loops[vIdx].rampDown(50::ms);
		50::ms => now;
		loops[vIdx].clear();
		loops[vIdx].rate(1.0);
		loops[vIdx].playPos(0::ms);
		loops[vIdx].loop(0);
		
		//begin ramping recording up (fade in) (so we can hear signal while we record)
		5::ms => loops[vIdx].recRamp;
		loops[vIdx].rampUp(5::ms);
		loops[vIdx].record(1);
		now => startTime;
		(now + maxDur) - startTime => loops[vIdx].loopEndRec;
	}
	
	//stop recording current loop
	fun void stopRec(int vIdx){
		//number of ms for fade out
		3 => int fade;
		//set end point of loop recording
		(now + fade::ms) - startTime => loops[vIdx].loopEndRec;
		fade::ms => loops[vIdx].recRamp;
		//fade out
		fade::ms => now;
		//get total length of loop
		now-startTime => durs[vIdx];
		//turn off recording
		loops[vIdx].record(0);
		//set end time
		loops[vIdx].loopEnd(voices[vIdx],durs[vIdx]);
		//start bi-directional looping playback
		loops[vIdx].play(1);
		loops[vIdx].loop(1);
		//loops[vIdx].bi(1);
		//<<<"stopping recording.  loop len: " + (now - startTime)/(1::samp)>>>;
	}
	
	
	//individual loop start method (called externally)
	fun void playLoop(int vIdx, dur fade){
		//play new voice
		//<<<"playing " + vIdx>>>;
		loops[vIdx].rampUp(fade);
		loops[vIdx].playPos(0::samp);
		loops[vIdx].play(1);
		loops[vIdx].loop(1);		
	}
	
	//individual loop stop method (called externally)
	fun void stopLoop(int vIdx){
		loops[vIdx].play(0);
		loops[vIdx].loop(0);
	}
	
	
	//clear set of loops. fade out, stop playback, and empty buffers
	fun void clearSet(int indices[]){
		for (int i; i < indices.size(); i++){
			loops[i].rampDown(50::ms);
			50::ms => now;
			loops[indices[i]].clear();
			loops[indices[i]].loop(0);
			loops[indices[i]].play(0);
		}
	}
	
	//clear all loops (fade out, stop playback, and empty buffer)
	fun void clearAll(){
		for (int i; i < numLoops; i++){
			loops[i].rampDown(50::ms);
			50::ms => now;
			
			loops[i].clear();
			loops[i].loop(0);
			loops[i].play(0);
			loops[i].bi(0);
			loops[i].rate(1);
		}
	}
	
	
	
	//set bidirectional loop playback for specific loop
	fun void setBi(int vIdx,int bi){
		//make sure vIdx falls in the appropriate range
		vIdx % (numLoops-1) => vIdx;
		loops[vIdx].bi(bi);
	}
	
	//global bidirectional playback mode
	fun void setBi(int bi){
		if (bi == 1){
			for (int i; i < numLoops; i++){
				loops[i].bi(bi);
				loops[i].bi() => bis[i];
			}
		}else{
			for (int i; i < numLoops; i++){
				0 => loops[i].bi;
			}
		}
	}
	
	//loop level control
	fun void setGain(int vIdx,float val){
		//make sure vIdx falls in the appropriate range
		vIdx % (numLoops-1) => vIdx;
		//make sure gain falls into appropriate range
		checkRange(val,0.0,0.95) => val;
		//don't change until we are close to current value
		softCheck(gains[vIdx].gain(),val,sThresh) => int ok;
		if (ok == 1){
			val => gains[vIdx].gain;
		}
	}
	
	
	//global level
	fun void setGain(float val){
		checkRange(val,0.0,0.95) => val;
		softCheck(dry.gain(), val, sThresh) => int ok;
		if (ok == 1){
			val => dry.gain;
		}
	}
	
	//loop send (wet gain)
	fun void send(int indices[],float wetGain){
		checkRange(wetGain,0.0,0.9) => wetGain;
		for (int i; i < indices.size(); i++){
			softCheck(wetGains[indices[i]].gain(), wetGain, sThresh) => int ok;
			if (ok == 1){
				wetGain => wetGains[indices[i]].gain;
			}
		}
	}
	
	//global loop send (all wet gains)
	fun void send(int vIdx,float wetGain){
		checkRange(wetGain,0.0,0.8) => wetGain;
		softCheck(wetGains[vIdx].gain(), wetGain, sThresh) => int ok;
		if (ok == 1){
			wetGain => wetGains[vIdx].gain;
		}
	}
	
	
	//global send
	fun void send(float wetGain){
		checkRange(wetGain,0.0,0.95) => wetGain;
		softCheck(wet.gain(), wetGain, sThresh) => int ok;
		if (ok == 1){
			wetGain => wet.gain;
		}
	}
	
	//loop playrate
	//modify playback rate of one or more loops
	fun void setRate(int indices[],float newRate){
		checkRange(newRate,minRate,maxRate) => newRate;
		for (int i; i < indices.size(); i++){
			softCheck(loops[indices[i]].rate(),newRate,rsThresh) => int ok;
			if (ok == 1){
				newRate => loops[indices[i]].rate;
			}
		}
	}
	
	//modify playback rate of one loop
	fun void setRate(int vIdx,float newRate){
		checkRange(newRate,minRate,maxRate) => newRate;
		softCheck(loops[vIdx].rate(),newRate,rsThresh) => int ok;
		if (ok == 1){
			newRate => loops[vIdx].rate;
		}
	}
	
	//global playrate (set playback rate of all loops) (multiplier between -1 and 1)
	fun void setRate(float newRate){
		checkRange(newRate,minRate,maxRate) => newRate;
		softCheck(globalRate,newRate,0.05) => int ok;
		if (ok == 1){
			for (int i; i < numLoops; i++){
				newRate=> loops[i].rate;
			}
			newRate => globalRate;
		}
	}
	
	
	//solo single voice
	fun void soloChan(int vIdx){
		//turn off all loops except current one
		for (int i; i < numLoops; i++){
			if (i != vIdx){
				if (solos[i] != 1){
					loopEnvs[i].keyOff();
				}
			}
		}
		//make sure current loop is on
		loopEnvs[vIdx].keyOn();
		
		//set solo state of this loop
		1 => solos[vIdx];
	}
	
	//unsolo single loop
	fun void unSoloChan(int vIdx){
		//check and see if other channels are still soloing.  
		0 => int sumSolos;
		for(int i; i < numLoops; i++){
			sumSolos + solos[i] => sumSolos;
		}
		
		//if other channels are soloing, only turn off current channel
		//otherwise, turn on all other channels (b/c nobody is soloing anymore)
		if (sumSolos > 1){
			loopEnvs[vIdx].keyOff();
		}else{
			for (int i; i < numLoops; i++){
				if (i != vIdx){
					loopEnvs[i].keyOn();
				}
			}
		}
		//update solo status of this voice
		0 => solos[vIdx];
	}
	
	//unsolo all items
	fun void unSoloAll(){
		for(int i; i < numLoops; i++){
			loopEnvs[i].keyOn();
		}
	}
	
	
	//pseudo soft takeover implementation (prevent jumps in values)
	fun int softCheck(float curVal, float targetVal, float threshDiff){
		if (Std.fabs(targetVal - curVal) < threshDiff){
			return 1;
		}else{
			return 0;
		}
	}
	
	
	//make sure value intended to fall between 0 - 1 falls within this range  (0.95 max 
	//since we're using this mostly with gain values)
	fun float checkRange(float val,float min, float max){
		if (val > max){
			max => val;
		}else if(val < min){
			min => val;
		}
		return val;
	}
	
	//getter function for total number of loops
	fun int getNumLoops(){
		return numLoops;
	}
	
	//set gain of direct output
	fun void setDirectGain(float val){
		checkRange(val,0.0,0.95) => val;
		softCheck(direct.gain(), val, sThresh) => int ok;
		if (ok == 1){
			val => direct.gain;
		}
	}
	
	//turn all filters on (pass along to filter spat object from midi control)
	fun void allOn(){
		fs.allOnOff(1,Std.rand2f(3,5)::second);
	}
	
	//turn all filters on (pass along to filter spat object from midi control)	
	fun void allOff(){
		fs.allOnOff(0,1::second);
	}
	
	
	
}